/**
Credits for the colors and font selection go to the Twitter Bootstrap framework.
*/


body {
    color: rgb(51, 51, 51);
    font-size: 10pt;
    padding-top: 10px;
    font-family: "Helvetica Neue",Helvetica,Arial,sans-serif;
}

h1 {
    font-size: 150%;
}

h2 {
    font-size: 140%;
}

h3 {
    font-size: 130%;
}

h4 {
    font-size: 120%;
}

h5 {
    font-size: 110%;
}

h6 {
    font-size: 100%;
}

div#attachments {
    display: inline-block;
    color: gray;
    border-width: 1px;
    border-style: solid;
    border-color: white;
    border-radius: 4px 4px 4px 4px;
    padding: 0px;
}

div#attachments dt {
    margin-top: 2px;
    margin-bottom: 2px;
}

div#attachments dd p {
    margin-top: 2px;
    margin-bottom: 2px;
}

div#attachments :target dt {
    font-weight: bold;
}

div#attachments :target a {
    color: rgb(70, 136, 71);
}


h1.title {
    text-align: center;
    font-size: 180%;
}

div.document {
    position: relative;
    background: white;
    max-width: 800px;
    margin: auto;
    padding: 20px;
    border: 1px solid rgb(221, 221, 221);
    border-radius: 4px 4px 4px 4px;
}

div.document:after {
    content: "snakemake report";
    position: absolute;
    top: -1px;
    right: -1px;
    padding: 3px 7px;
    background-color: #f5f5f5;
    border: 1px solid rgb(221, 221, 221);
    color: #9da0a4;
    font-weight: bold;
    font-size: 12pt;
    border-radius: 0 0 0 4px;
}

div.document p {
    text-align: justify;
}

div#metadata {
    text-align: right;
}

table.docutils {
    border: none;
    border-collapse: collapse;
    border-top: 2px solid gray;
    border-bottom: 2px solid gray;
    text-align: center;
}

table.docutils th {
    border: none;
    border-top: 2px solid gray;
    border-bottom: 2px solid gray;
    padding: 5px;
}

table.docutils td {
    border: none;
    padding: 5px;
}

table.docutils th:last-child, td:last-child {
    text-align: left;
}

table.docutils th:first-child, td:first-child {
    text-align: right;
}

table.docutils th:only-child, td:only-child {
    text-align: center;
}

table.docutils.footnote {
    border: none;
    text-align: left;
}

a {
    color: rgb(0, 136, 204);
    text-decoration: none;
}

a:hover {
    color: rgb(0, 85, 128);
    text-decoration: underline;
}


div.figure {
    margin-left: 2em;
    margin-right: 2em;
}

img {
    max-width: 100%;
}

p.caption {
    font-style: italic;
}
