
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-06-30T15:19:34+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "531ef4a6b962b06ef865ba34ea6e952a7d4b3a2b",
 "version": "7.8.5"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
