package_management = "https://snakemake.readthedocs.io/en/latest/snakefiles/deployment.html#integrated-package-management"
containers = "https://snakemake.readthedocs.io/en/latest/snakefiles/deployment.html#running-jobs-in-containers"
rules = "https://snakemake.readthedocs.io/en/latest/snakefiles/rules.html#rules"
external_scripts = (
    "https://snakemake.readthedocs.io/en/latest/snakefiles/rules.html#external-scripts"
)
notebooks = "https://snakemake.readthedocs.io/en/latest/snakefiles/rules.html#jupyter-notebook-integration"
checkpoints = "https://snakemake.readthedocs.io/en/latest/snakefiles/rules.html#data-dependent-conditional-execution"
params = "https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#non-file-parameters-for-rules"
log = "https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#log-files"
input_functions = "https://snakemake.readthedocs.io/en/stable/tutorial/advanced.html#tutorial-input-functions"
envvars = "https://snakemake.readthedocs.io/en/latest/snakefiles/configuration.html#environment-variables"
includes = (
    "https://snakemake.readthedocs.io/en/latest/snakefiles/modularization.html#includes"
)
config = "https://snakemake.readthedocs.io/en/latest/snakefiles/configuration.html#configuration"
